#include "load_denik.h"

/*
 * load_denik.c [book of original entry, cesky ucetni denik]
 *
 * program to load journal entries into the MDEBS database
 * Takes one argument: name of datafile to load entries from
 */

FILE *fp, *logfp, *rejfp;
int startdate, enddate;         /* starting and ending dates of fiscal year *
				 * in yyyymmdd form                         */

/* various buffers to hold one journal entry prior to insert into Postgres */
char *rec[10];             /* array of buffers to hold raw journal
		            * entry strings as read from datafile */
char *entrydate;
char *descr;
char *mdacct[10], *mdanal[10], *dacct[10], *danal[10];
int mdamt[10], damt[10];

/* buffer to hold max_cislo from table 'porad' */
int porcislo;

void main(int argc, char *argv[])
{
  short i, j, k;
  char *logfilename, *rejfilename;

  /* Control number of command line arguments */
  if (argc < 2) {
    fprintf(stderr, "Error: Too few arguments: must provide name of datafile.\n");
    exit(0);
  }

  /* note that argv[1] must be the name of the datafile containing 
   * journal entries to be processed   */
  printf("Opening datafile %s...\n", argv[1]);
  if ((fp = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, 
                "Fatal Error: Could not open %s for reading.\n", argv[1]);
    exit(0);
  }

  /* Initialize the filename variables */
  logfilename = malloc(strlen(argv[1])+5);
  logfilename = strcpy(logfilename, argv[1]);
  logfilename = strcat(logfilename, ".log");
  rejfilename = malloc(strlen(argv[1])+5);
  rejfilename = strcpy(rejfilename, argv[1]);
  rejfilename = strcat(rejfilename, ".rej");

  /* Open the log file */
  if ((logfp = fopen(logfilename, "a")) == NULL) {
    fprintf(stderr,
	    "Fatal Error: Could not open %s for appending.\n", logfilename);
    exit(0);
  } else {
    fprintf(logfp, "==============================================\n");
    fprintf(logfp, "load_denik RUNNING AT %s\n", theTimeIs());
    fprintf(logfp, "==============================================\n");
    fflush(logfp);
  }

  /* Allocate memory to global variables */
  entrydate = malloc(12);
  descr = malloc(61);
  for (i = 0; i <= 9; i++) {
    rec[i] = malloc(81);
    mdacct[i] = malloc(4);
    mdanal[i] = malloc(3);
    dacct[i] = malloc(4);
    danal[i] = malloc(3);
  }

  fprintf(logfp, "Initializing the backend...\n");
  fflush(logfp);
  initBackend();                /* initialize the backend connection */

  /* this function call initializes the startdate and enddate variables
   * (see above)
   */
  fprintf(logfp, "Querying fiscal year... ");
  fflush(logfp);
  queryFiscalYear(&startdate, &enddate);
  fprintf(logfp, "OK (%d - %d)\n", startdate, enddate);
  fflush(logfp);

  /* this is the main loop of the program
   * it is run once for each journal entry in the datafile */
  j = 0;
  while ((i = readInTrans()) != 0) {
    j++;
    if (i ==  -1) { /* bad transaction encountered -- reject it */
      printf("Rejecting transaction no. %d\n", j);
      fprintf(logfp, "Rejecting the transaction.\n");
      fflush(logfp);
      if ((rejfp = fopen(rejfilename, "a")) == NULL) {
	fprintf(stderr,
	    "Fatal Error: Could not open %s for appending.\n", rejfilename);
	exit(0);
      } else {
	for (k=0; *(rec[k]) != '\0'; k++) {
	  fprintf(rejfp, "%s", rec[k]);
	  fflush(rejfp);
	  if (k == 9) {
	    fprintf(logfp, "Reject loop overflow.\n");
	    fflush(logfp);
	    break;
	  }
	}
	fprintf(rejfp, "\n");
	fflush(rejfp);
      }
    } /* end of code to reject bad transaction */ else {
      printf("Transaction no. %d was read and seems OK...\n", j);
      writeOutTrans(i);
      printf("   ...written to database.\n");
    }
  }

  printf("Done! Thank you for your patronage.\n");
  fclose(fp);
  fclose(logfp);
  exit_nicely();
}

/*
 * queryPorCislo()
 *
 * Function to determine the poradove cislo for a particular date (in
 * pg format) */
int queryPorCislo(char *thisdate)
{
  char miscbuff[81];
  int resultcode, noofrecs;
  int retval;

  /* fprintf(logfp, "Querying poradove cislo...\n");
   * fflush(logfp); */

  sprintf(miscbuff, "SELECT max_cislo FROM porad where ent_date='%s';", thisdate);
  resultcode = runQuery(miscbuff);

  /* fprintf(logfp, "Query result code was %d.\n", resultcode);
   * fflush(logfp); */

  if (resultcode == 3) {                /* this is the proper result code */
    noofrecs = numberReturned();

    /* fprintf(logfp, "Number of records returned was %d.\n", noofrecs);
     * fflush(logfp); */

    if (noofrecs > 1) {
      fprintf(stderr, "Error: Corrupted database: porad contains more than one record for this date.\n");
      exit_nicely();
    } else
      if (noofrecs == 1) {
	/* code to get the fields from the one record returned */
	retval = atoi(getField("max_cislo"));

	/* fprintf(logfp, "Max_cislo for %s is %d.\n", thisdate, retval);
	 * fflush(logfp); */

      } else {  /* query returned no records */
        retval = 0;
      }
  } else {
    fprintf(stderr, "Error: Query returned improper result code (%d).\n", resultcode);
    exit_nicely();
  }
  return retval;
}

/*
 * writeOutTrans()
 *
 * Function to insert the transaction into the relevant postgres
 * database tables */
void writeOutTrans(short i)
{
  char miscbuff[261];
  int resultcode;
  int j;

  fprintf(logfp, "Data looks OK; recording...\n");
  fflush(logfp);

  /* fire up a transaction */
  startTransaction();

  /* code to insert into table 'porad' */
  if (++porcislo > 1) {   /* first we have to delete the current value */
    sprintf(miscbuff, "DELETE FROM porad where ent_date='%s';", entrydate);
    resultcode = runQueryInTrans(miscbuff);
    if (resultcode == 2) {
      fprintf(logfp, "Delete from table 'porad' successful.\n");
      fflush(logfp);
    } else {
      fprintf(stderr, "'porad' delete command returned (%d), which is not what I expected.\n", resultcode);
      fprintf(logfp, "'porad' delete command returned (%d), which is not what I expected.\n", resultcode);
      fflush(logfp);
      rollbackTransaction();
      exit_nicely();
    }
  }

  sprintf(miscbuff, "INSERT INTO porad VALUES ('%s', %d);", entrydate , porcislo);
  resultcode = runQueryInTrans(miscbuff);
  if (resultcode == 2) {
    fprintf(logfp, "Insert into table 'porad' successful.\n");
    fflush(logfp);
  } else {
    fprintf(stderr, "'porad' insert command returned (%d), which is not what I expected.\n", resultcode);
    fprintf(logfp, "'porad' insert command returned (%d), which is not what I expected.\n", resultcode);
    fflush(logfp);
    rollbackTransaction();
    exit_nicely();
  }
  
  /* code to insert into table 'descs' */
  sprintf(miscbuff, "INSERT INTO descs VALUES ('%s', %d, '%s');", entrydate , porcislo, descr);
  resultcode = runQueryInTrans(miscbuff);
  if (resultcode == 2) {
    fprintf(logfp, "Insert into table 'descs' successful.\n");
    fflush(logfp);
  } else {
    fprintf(stderr, "'descs' insert command returned (%d), which is not what I expected.\n", resultcode);
    fprintf(logfp, "'descs' insert command returned (%d), which is not what I expected.\n", resultcode);
    fflush(logfp);
    rollbackTransaction();
    exit_nicely();
  }
  
  /* code to insert into table 'denik' */
  for (j = 1; j < i; j++) {
    sprintf(miscbuff, "INSERT INTO denik VALUES ('%s', %d, '%s', '%s', %d, '%s', '%s', %d);", entrydate , porcislo, mdacct[j], mdanal[j], mdamt[j], dacct[j], danal[j], damt[j]);
    resultcode = runQueryInTrans(miscbuff);
    if (resultcode == 2) {
      fprintf(logfp, "Insert into table 'denik' successful.\n");
      fflush(logfp);
    } else {
      fprintf(stderr, "'denik' insert command returned (%d), which is not what I expected.\n", resultcode);
      fprintf(logfp, "'denik' insert command returned (%d), which is not what I expected.\n", resultcode);
      fflush(logfp);
      rollbackTransaction();
      exit_nicely();
    }
  }

  /* if we're still alive at this point, the transaction was successful */
  endTransaction();

  fprintf(logfp, "Transaction was successfully recorded.\n\n");

  return;
}
  
/*
 * readInTrans()
 *
 * Function to read in one transaction from datafile into global
 * buffers.  Returns 0 on end of datafile; number of non-date lines in
 * transaction otherwise */
int readInTrans()
{
  int i, j;     /* i is counter for PART I., j is counter for PART II. */
  int nathandate;
  char *recprime[11];
  char *minibuff[2];
  int  mdcounter, dcounter;
  char *acctnam;
  
  /*
   * PART I.
   *
   * READ THE RAW DATA INTO THE REC[] ARRAY
   */
  fprintf(logfp, "Reading in raw data from datafile...\n");
  fflush(logfp);
  i = 0;
  do {
    fprintf(logfp, "i = %d\n", i);
    fflush(logfp);
    if (i == 10) {
      fprintf(stderr, "Datafile error: entry with more than 9 or 10 lines encountered!\n");
      exit_nicely();
    }
    if (fgets(rec[i], 81, fp) == NULL) { /* in other words, if we've
					    reached the end of the
					    file */
      strcpy(rec[i], " ");
      fprintf(logfp, "End of file encountered.\n");
      fflush(logfp);
    } else {
      if (countNonWhiteSpace(rec[i]) > 0) {
	fprintf(logfp, "%s", rec[i]);
	fflush(logfp);
      }
    }
  } while (countNonWhiteSpace(rec[i++]) > 0); /* in other words, while
						 the string that is
						 read from the datafile
						 contains characters
						 other than space, tab,
						 and newline */

  *(rec[--i]) = '\0';

  if (i < 3) { 
    /* i _should_ be zero if there are no more transactions left
     * in the file (otherwise the last 'transaction' is incomplete);
     * consider printing out a warning here if it is not */
    if (i != 0) {
      fprintf(logfp, "Warning: End of file encountered before end of transaction.\n");
      fflush(logfp);
    }
    return 0;
  } else {
    fprintf(logfp, "Transaction with total of %d lines.\n", i);
    fflush(logfp);
  }

  /*
   * PART II.
   *
   * PARSE THE RAW DATA INTO THE GLOBAL BUFFERS
   */

  /* Allocate memory for the recprime array: (i - 1) == number of lines
   * in transaction (not including blank line at end) */
  for (j = 0; *(rec[j]) != '\0'; j++) recprime[j] = malloc(81);

  /* Make copy of the rec array */
  for (j = 0; *(rec[j]) != '\0'; j++) strcpy(recprime[j], rec[j]);

  /* Allocate memory for the minibuff array */
  minibuff[0] = malloc(81);
  minibuff[1] = malloc(81);

  /* here we attempt to parse the first line of the transaction, which
   * is supposed to contain the date and the journal entry
   * description.  */
  nathandate = atoi(strtok(recprime[0], " \t\n"));
  descr = strtok(NULL, "\n\0");

  /* This is as good a time to check the validity of the date as
     any. */
  if ((nathandate < startdate) || (nathandate > enddate)) {
    fprintf(stderr, "Warning: Journal entry's date falls outside of fiscal year.\n");
    fprintf(logfp, "Transaction date out of fiscal year range encounterd.\n");
    fflush(logfp);
    return -1;
  }
  entrydate = pgConvertDate(entrydate, nathandate);

  /* Query the poradove cislo table to find the greatest yet-used
   * poradove cislo for the date of this journal entry. The poradove
   * cislo of the current journal entry will be one greater than
   * this */
  fprintf(logfp, "Running queryPorCislo()...\n");
  fflush(logfp);
  porcislo = queryPorCislo(entrydate);
  fprintf(logfp, "  ... queryPorCislo() returned OK\n");
  fflush(logfp);
  
  for (j = 1; *(rec[j]) != '\0'; j++) {
    fprintf(logfp, "recprime[%d] = %s", j, recprime[j]);
    fflush(logfp);
  }

  fprintf(logfp, "Parsing rows 1 through %d of recprime[]...\n", (i - 1));
  fflush(logfp);

  /* Parse the rest of the recprime array */
  mdcounter = 0;
  dcounter = 0;
  for (j = 1; *(rec[j]) != '\0'; j++) {

    /*
     * log for loop deciphering
     */
    fprintf(logfp, "j = %d\n", j);
    fflush(logfp);

    /*
     * check for correct number of fields before parsing
     */
    fprintf(logfp, "The line has %d tokens.\n", strntok(recprime[j]));
    fflush(logfp);
    if (strntok(recprime[j]) != 4) {
      fprintf(stderr, "Error: Line in journal entry does not contain enough fields.\n");
      fprintf(logfp, "Error: Line in journal entry does not contain enough fields.\n");
      fprintf(logfp, "The line was: \"%s\"", rec[j]);
      fflush(logfp);
      exit_nicely();
    }

    /* 
     * get the tokens
     */
    minibuff[0] = strtok(recprime[j], " \t\n\0");
    minibuff[1] = strtok(NULL, " \t\n\0");
    mdamt[j] = atoi(strtok(NULL, " \t\n\0"));
    damt[j] = atoi(strtok(NULL, "\n\0"));

    /*
     * check the length of the acct and anal portions
     */
    if ((strlen(minibuff[0]) != 3) || (strlen(minibuff[1]) != 2)) {
      fprintf(stderr, "Warning: Account number is not in format XXX YY\n");
      fprintf(logfp, "Warning: Account number is not in format XXX YY\n");
      fflush(logfp);
      return -1;
    }

    /* if mdamt is zero, then the account number given by
     * minibuff[0] and minibuff [1] is a dacct/danal.  If the damt
     * is zero, then it's the other way around.  If both are not
     * zero, then we have a fatal error */
    if (mdamt[j] == 0) {
      strcpy(dacct[j], minibuff[0]);
      strcpy(danal[j], minibuff[1]);
      strcpy(mdacct[j], "000");
      strcpy(mdanal[j], "--");
    } else
      if (damt[j] == 0) {
	strcpy(mdacct[j], minibuff[0]);
	strcpy(mdanal[j], minibuff[1]);
	strcpy(dacct[j], "000");
	strcpy(danal[j], "--");
      } else {
	fprintf(stderr, "Error: Journal entry contains line without zero value.\n");
	fprintf(logfp, "Error: Journal entry contains line without zero value.\n");
	fflush(logfp);
	return -1;
      }
    
      /* increment the counters */
      mdcounter = mdcounter + mdamt[j];
      dcounter = dcounter + damt[j];
  }

  /* If the counters are not equal, then the journal entry does not
   * add up, which is a fatal error. */
  if (mdcounter != dcounter) {
    fprintf(stderr, "Error: Unbalanced journal entry in datafile!\n");
    fprintf(logfp, "Error: Unbalanced journal entry in datafile!\n");
    fflush(logfp);
    return -1;
  }

  /* Determine if account numbers are valid */
  acctnam = malloc(61);
  for (j = 1; *(rec[j]) != '\0'; j++) {
    if (strcmp(mdacct[j], "000") != 0)
      if ((acctnam = acctQuery(acctnam, mdacct[j], mdanal[j])) == NULL) {
	fprintf(stderr, "Warning: Account %s %s is not in the ucetni osnova!\n", mdacct[j], mdanal[j]);
	fprintf(logfp, "Warning: Account %s %s is not in the ucetni osnova!\n", mdacct[j], mdanal[j]);
	fflush(logfp);
	return -1;
      }
      else {
	fprintf(logfp, "Found %s %s: %s\n", mdacct[j], mdanal[j], acctnam);
	fflush(logfp);
      }
    if (strcmp(dacct[j], "000") != 0)
      if ((acctnam = acctQuery(acctnam, dacct[j], danal[j])) == NULL) {
	fprintf(stderr, "Warning: Account %s %s is not in the ucetni osnova!\n", dacct[j], danal[j]);
	fprintf(logfp, "Warning: Account %s %s is not in the ucetni osnova!\n", dacct[j], danal[j]);
	fflush(logfp);
	return -1;
      }
      else {
	fprintf(logfp, "Found %s %s: %s\n", dacct[j], danal[j], acctnam);
	fflush(logfp);
      }
  }

  return j;
}


